c-----------------------------------------------------------
c    'mgfas2d' and supporting routines constitute a complete 
c    FAS multi-grid solver for
c
c        u_xx + u_yy + G(u) = rhs(x,y)       (1)
c
c     on the unit square with homogenous Dirichlet boundary
c     conditions.  Literature (particularly Brandt) should be
c     consulted for treatment of Neumann and mixed boundary
c     conditions.
c
c     Currently program solves (1) with
c
c        G(u) = sigma * u*u
c
c     and 'rhs' specified so that
c
c        u = sin(Pi lx x) sin(Pi ly y)
c
c     for some integers, 'lx' and 'ly' defined in 
c     'mgprob.inc'.
c-----------------------------------------------------------
      program      mgfas2d

      implicit     none

c-----------------------------------------------------------
c     Functions called
c-----------------------------------------------------------
      integer      iargc,      i4arg
      real*8       r8arg,      dmnrm2

c-----------------------------------------------------------
c     Command-line arguments
c-----------------------------------------------------------
      integer      level,      nvcycle,    preswp,    pstswp
      integer      lmin
c  Physical Radius of mask on finest grid.
c  Note that negative value turns excision off.
      real*8       mrad        
c  Displacement in x-direction of each exicision region
      real*8       mdisp_x

c-----------------------------------------------------------
c     'mg.inc' contains the main 'data structure' used 
c     in this implementation.
c-----------------------------------------------------------
      include     'mg.inc'

c-----------------------------------------------------------
c     'mgprob.inc' contains problem-specific parameters
c     including command-line arguments 'sigma' and 
c     'gft_trace'.
c-----------------------------------------------------------
      include     'mgprob.inc'

c-----------------------------------------------------------
c     Locals.
c-----------------------------------------------------------
      integer      l,          cycle

c-----------------------------------------------------------
c     Argument parsing.
c
c     level:     Finest level solution requested.  Full MG 
c                process solves sequence of problems 
c                l = lmin ... level (n_l = 2^l+1 pts per side)
c     lmin:      Lowest level used in multigrid solve
c     mrad:      "Radius" of excision mask
c     mdisp_x:   Dispacement (in x-direction) of each of two
c                excision regions (e.g. two circles)
c     sigma:     Controls nonlinearity in PDE (1)
c     nvcycle:   # of vcycles to apply to each problem
c     preswp:    # of relaxation sweeps before a CGC 
c     pstswp:    # of relaxation sweeps after a CGC
c     gft_trace: If present, enables RNPL-style .hdf 
c                output of solution estimate 'u' after 
c                each relaxation sweep.
c-----------------------------------------------------------
      if( iargc() .lt. 1 ) go to 900
      level     = i4arg(1,-1)
      if( level .le. 0 ) go to 900
      if( level .gt. lmax ) then
         write(0,*) 'mgfas2d: Specified level is too high'
         write(0,*) 'mgfas2d: Current maximum level: ', lmax
         stop
      end if
      lmin      = i4arg(2,1)
      if (level .lt. lmin) then
         write(0,*) 'mgfas2d: Error: level < lmin'
         stop
      end if

      mrad      = r8arg(3,-1.0d0)
      mdisp_x   = r8arg(4,0.0d0)
      sigma     = r8arg(5,0.0d0)

      nvcycle   = i4arg(6,1)
      preswp    = i4arg(7,1)
      pstswp    = i4arg(8,1)
      gft_trace = iargc() .gt. 8

      if( gft_trace ) write(0,*) 'gft tracing enabled'

c-----------------------------------------------------------
c     Initialize data structure.  All grid function arrays
c     are allocated from global array 'q' (see 'mg.inc')
c-----------------------------------------------------------
      call gsinit(level)

c-----------------------------------------------------------
c     Set up the mask
c-----------------------------------------------------------
      call minit(level,lmin,mrad,mdisp_x)
 
c-----------------------------------------------------------
c     Solve series of problems from l = lmin through 
c     l = level.
c-----------------------------------------------------------
      do l = lmin , level
         if( l .eq. lmin ) then
c-----------------------------------------------------------
c           Initialize level 1 solution (coarsest) with 0.
c-----------------------------------------------------------
            call dmls  ( q(u(l)), 0.0d0, nx(l), ny(l) )
            call exonm( q(u(l)), qm(mask(l)), h(l),
     &          nx(l), ny(l), m_ex )
         else
c-----------------------------------------------------------
c           Initialize level l soln via linear interpolation
c           of level l-1 soln.
c-----------------------------------------------------------
            call exonm( q(u(l-1)), qm(mask(l-1)), h(l-1),
     &          nx(l-1), ny(l-1), m_ex )
            call dmlint ( q(u(l-1)), q(u(l)), 
     &                    nx(l-1), ny(l-1) )
            call exonm( q(u(l)), qm(mask(l)), h(l),
     &          nx(l), ny(l), m_ex )
         end if
c-----------------------------------------------------------
c        Initialize 'rhs' (rhs = f).
c-----------------------------------------------------------
         call clcrhs2   ( q(rhs(l)), h(l), nx(l), ny(l) )
c-----------------------------------------------------------
c        "Solve" level l problem by applying cycle 
c        (preswp,pstswp) V-cycles.
c-----------------------------------------------------------
         do cycle = 1 , nvcycle
            call vcycle(l,lmin,cycle,preswp,pstswp)
         end do

         if (l .gt. lmin) then
            call calctau(l)
            write(0,*) 'l, || tau_h^2h|| = ', l,
     &                  dmnrm2( q(tau(l-1)), nx(l-1), ny(l-1) )
         endif

c-----------------------------------------------------------
c        Compute norm of error on finest gird and output to 
c        standard error.
c-----------------------------------------------------------
         call clcu2 ( q(t1(l)), h(l), nx(l), ny(l) )
         call dmms  ( q(t1(l)), q(u(l)), q(t1(l)), 
     &                nx(l), ny(l) )

         write(0,1000) l,dmnrm2( q(t1(l)), nx(l), ny(l) )
         write(*,1000) l,dmnrm2( q(t1(l)), nx(l), ny(l) )
1000     format (i4,e25.16)

      end do
      l = level

      call gftout(q(t1(l)),'err',nx(l),ny(l))

c-----------------------------------------------------------
c     Close RNPL-style .hdf files if option enabled.
c-----------------------------------------------------------
      if( gft_trace ) then
         call gft_close_all()
      end if

c-----------------------------------------------------------
c     Normal exit.
c-----------------------------------------------------------
      stop

c-----------------------------------------------------------
c     Usage exit.
c-----------------------------------------------------------
 900  continue
         write(0,*) 'usage: mgfas2d_ex <level> [<lmin> <mrad> '//
     &              '<mdisp_x> <sigma> '//
     &              '<nvcycle> <preswp> <pstswp> <gft_trace>]'
      stop

      end

c-----------------------------------------------------------
c     Performs one FAS V-cycle of level 'l' problem, using 
c     'preswp' pre-CGC and 'pstswp' pst-CGC sweeps.  
c     Supply cycle=N for Nth consecutive invocation of this 
c     routine on the same level.
c-----------------------------------------------------------
      subroutine vcycle(l,lmin,cycle,preswp,pstswp)
 
         implicit       none 
 
         include       'mg.inc'

         integer        l,        lmin
         integer        cycle,    preswp,   pstswp
 
         real*8         dmnrm2,   relaxc2
 
         integer        m,        sweep
         real*8         rresnrm

         logical        use_esr
         parameter    ( use_esr = .false. )
 
c-----------------------------------------------------------
c        Work from fine level to coarse level, performing
c        pre-CGC sweeps and intitiating coarse grid 
c        corrections.  Skipped if on coarsest level, l = lmin.
c-----------------------------------------------------------
         do m = l , lmin+1 , -1
c-----------------------------------------------------------
c           Pre-CGC relaxation sweeps.
c-----------------------------------------------------------
c            call exonm  ( q(u(m)), qm(mask(m)), h(m),
c     &                    nx(m), ny(m), m_ex )
            if( cycle .eq. 1  .or.  m .ne. l ) then
               do sweep = 1 , preswp

c               Relax an extra time in the "extra smoothing region"
                  if (use_esr)
     &               rresnrm = relaxc2(q(u(m)),q(rhs(m)),
     &                              nx(m),ny(m),h(m),
     &                              qm(mask(m)), m_esr, m_esr )
c               The "real" relaxation
                  rresnrm = relaxc2(q(u(m)),q(rhs(m)),
     &                              nx(m),ny(m),h(m),
     &                              qm(mask(m)), m_int, m_esr )
c            call exonm  ( q(u(m)), qm(mask(m)), h(m),
c     &                    nx(m), ny(m), m_ex )
                  call gftout(q(u(m)),'u',nx(m),ny(m))
                  call trace(m,'Pre-CGC',rresnrm)
               end do
            end if

c-----------------------------------------------------------
c           Compute truncation error estimate.
c
c               2h      2h 2h h     2h h  h
c           tau    :=  L  I  u  -  I  L  u
c               h         h        h
c-----------------------------------------------------------
            call calctau(m)
c            write(0,*) '|| tau_h^2h|| = ',
c     &                dmnrm2( q(tau(m-1)), nx(m-1), ny(m-1) )
c-----------------------------------------------------------

c
c               2h       2h    h       2h
c           rhs     :=  I   rhs   + tau
c                        h              h
c
c            2h         2h  h
c           u       := I   u
c                       h
c-----------------------------------------------------------
            call dmrshw ( q(rhs(m)), q(rhs(m-1)), nx(m-1), ny(m-1),
     &                    qm(mask(m)), qm(mask(m-1)),m_ex)
            call dmma   ( q(rhs(m-1)), q(tau(m-1)), q(rhs(m-1)),
     &                    nx(m-1), ny(m-1) )

            call dmrshw ( q(u(m)), q(u(m-1)), nx(m-1), ny(m-1),
     &                    qm(mask(m)), qm(mask(m-1)),m_ex )
c            call exonm( q(u(m-1)), qm(mask(m-1)), h(m-1),
c     &          nx(m-1), ny(m-1), m_ex )

c-----------------------------------------------------------
c           Continue loop to perform coarse-grid 
c           correction.
c-----------------------------------------------------------
         end do

c-----------------------------------------------------------
c        Solve coarse-grid problem using relaxation.
c        Note use of 'do while()'
c-----------------------------------------------------------
         rresnrm = 1.0d0 + epsi1
         sweep = 1
c         call exonm  ( q(u(lmin)), qm(mask(lmin)), h(lmin),
c     &                 nx(lmin), ny(lmin), m_ex )
         do while( rresnrm .gt. epsi1  .and. 
     &             sweep .le. maxsweep1 )
            if (use_esr)
     &          rresnrm = relaxc2( q(u(lmin)), q(rhs(lmin)),
     &                         nx(lmin), ny(lmin), h(lmin),
     &                         qm(mask(lmin)), m_esr, m_esr )
            rresnrm = relaxc2( q(u(lmin)), q(rhs(lmin)),
     &                         nx(lmin), ny(lmin), h(lmin),
     &                         qm(mask(lmin)), m_int, m_esr )
c            call exonm  ( q(u(lmin)), qm(mask(lmin)), h(lmin),
c     &                    nx(lmin), ny(lmin), m_ex )
            sweep = sweep + 1
            call gftout(q(u(lmin)),'u',nx(lmin),ny(lmin))
            call trace(lmin,'Coarse grid',rresnrm)
         end do
         if( rresnrm .gt. epsi1 ) then
            write(0,*) '*** WARNING vcycle: Coarse grid '//
     &                 'relaxation failed.'
         end if

c-----------------------------------------------------------
c        Now work from coarse to fine level, updating 
c        unknowns and performing post-CGC relaxtion 
c        sweeps.  Skipped if l = lmin.
c-----------------------------------------------------------
         do m = lmin+1 , l
c-----------------------------------------------------------
c           Linearly interpolate correction and update 
c           level m unknown.
c
c            h       h    h    2h     2h  h
c           u    := u  + I  ( u   -  I   u  )
c                         2h          h
c-----------------------------------------------------------

            call dmrshw ( q(u(m)), q(t1(m-1)), nx(m-1), ny(m-1),
     &                    qm(mask(m)), qm(mask(m-1)),m_ex )
            call dmms   ( q(u(m-1)), q(t1(m-1)), q(t1(m-1)),
     &                    nx(m-1), ny(m-1) )
            call dmlint ( q(t1(m-1)), q(t2(m)),
     &                    nx(m-1),ny(m-1) )
            call dmma   ( q(u(m)), q(t2(m)), q(u(m)),
     &                    nx(m),ny(m) )

c-----------------------------------------------------------
c           Post-CGC relaxation sweeps.
c-----------------------------------------------------------
            if (m .eq. l)
     &        call exonm  ( q(u(m)), qm(mask(m)), h(m),
     &                      nx(m), ny(m), m_ex )
            do sweep = 1 , pstswp
               if (use_esr)
     &             rresnrm = relaxc2 ( q(u(m)), q(rhs(m)),
     &                             nx(m),ny(m),h(m), 
     &                             qm(mask(m)), m_esr, m_esr )
               rresnrm = relaxc2 ( q(u(m)), q(rhs(m)),
     &                             nx(m),ny(m),h(m), 
     &                             qm(mask(m)), m_int, m_esr )
               call gftout(q(u(m)),'u',nx(m),ny(m))
               call trace(m,'Post-CGC',rresnrm)
            end do
          end do

c-----------------------------------------------------------
c         End of the V-cycle.
c-----------------------------------------------------------
         write(0,*)

         return
      end
